# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p


class TronAccountUpdateContract(p.MessageType):

    def __init__(
        self,
        account_name: str = None,
    ) -> None:
        self.account_name = account_name

    @classmethod
    def get_fields(cls):
        return {
            1: ('account_name', p.UnicodeType, 0),
        }
