# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p


class TronFreezeBalanceContract(p.MessageType):

    def __init__(
        self,
        frozen_balance: int = None,
        frozen_duration: int = None,
    ) -> None:
        self.frozen_balance = frozen_balance
        self.frozen_duration = frozen_duration

    @classmethod
    def get_fields(cls):
        return {
            1: ('frozen_balance', p.UVarintType, 0),
            2: ('frozen_duration', p.UVarintType, 0),
        }
